/**
 *  '$RCSfile: AccessionNumber.java,v $'
 *    Purpose: A class that gets Accession Number, check for uniqueness
 *             and register it into db
 *  Copyright: 2000 Regents of the University of California and the
 *             National Center for Ecological Analysis and Synthesis
 *    Authors: Jivka Bojilova, Matt Jones
 *    Release: 1.0.3
 *
 *   '$Author: bojilova $'
 *     '$Date: 2001/06/12 17:15:49 $'
 * '$Revision: 1.25 $'
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package edu.ucsb.nceas.metacat;

import java.net.*;
import java.sql.*;

import edu.ucsb.nceas.dbadapter.AbstractDatabase;

/**
 * (on insert of XML document)
 * Generates a unique Accession Number or if provided check it 
 * for uniqueness and register it into the db connection 
 * (on update or delete of XML document)
 * Check for existance of provided Accession Number
 * 
 */
public class AccessionNumber  {
  
  private static final AbstractDatabase dbAdapter = MetaCatUtil.dbAdapter;

  private Connection conn = null;
  private String sitecode = null;
  private String sep = null;
  private String docid = null;
  private String rev = null;
    
  /** 
   * Construct an AccessionNumber 
   */
  public AccessionNumber () 
         throws SQLException, ClassNotFoundException  
  {
    // Open a connection to the database from here
    this((new MetaCatUtil()).openDBConnection());
  }  

  /** 
   * Construct an AccessionNumber
   *
   * @param conn the db connection to read from and write Accession# to
   */
  public AccessionNumber (Connection conn) 
  {
    this.sitecode = MetaCatUtil.getOption("sitecode");
    this.sep = MetaCatUtil.getOption("accNumSeparator");
    this.conn = conn;
  }

  /** NEW - WHEN CLIENT ALWAYS PROVIDE ACCESSION NUMBER INCLUDING REV IN IT
   * Construct an AccessionNumber
   *
   * @param conn the db connection to read Accession number from
   * @param accnum the accession number to be checked for validness
   */
  public AccessionNumber (Connection conn, String accnum, String action) 
                         throws AccessionNumberException, SQLException 
  {
    this(conn);

    this.rev = null;
    this.docid = accnum;
    if ( accnum != null ) {
      int firstIndex = accnum.indexOf(this.sep);
      int lastIndex = accnum.lastIndexOf(this.sep);
      if ( firstIndex != lastIndex ) {
        //this docid contains a revision number
        this.rev = accnum.substring(lastIndex + 1);
        this.docid = accnum.substring(0, lastIndex);
      }
    }
    
    // INSERT
    if ( action.equals("INSERT")) {

      // check accession number for validness
      if ( docid == null ) {
        throw new AccessionNumberException("Accession number is required.");

      // rev is not provided; throw an exception to prevent the insertion
      } else if ( rev == null ) {
        throw new AccessionNumberException
                  ("Revision number is required and must be 1.");

      // docid is used; throw an exception to prevent the insertion
      } else if ( accNumberUsed(docid) ) {
        throw new AccessionNumberException
                  ("Accession number " + docid + " is already in use.");
                    
      // rev is <> 1; throw an exception to prevent the insertion
      } else if ( !rev.equals("1") ) {
        throw new AccessionNumberException("Revision number must be 1.");
      }
          
    // UPDATE or DELETE
    } else if ( action.equals("UPDATE") || action.equals("DELETE")) {
      String l_rev = "";

      // Accession# is not provided; throw an exception to prevent the action
      if ( docid == null ) {
        throw new AccessionNumberException("Accession number is required.");

      // rev is not provided; throw an exception to prevent the action
      } else if ( rev == null ) {
        throw new AccessionNumberException
                  ("Revision number is required.");

      // Accession# is not current (not in xml_documents); throw an exception
      } else if ( !accNumberIsCurrent(docid) ) {
        throw new AccessionNumberException
                  ("Document not found for Accession number " + docid);

      // Revision number is not the recent one; throw an exception
      } else if ( action.equals("UPDATE") &&
                  !rev.equals(l_rev = getNextRevision(docid)) ) {
        throw new AccessionNumberException
                  ("Next revision number must be "+ l_rev);

      // Revision number is not the recent one; throw an exception
      } else if ( action.equals("DELETE") && 
                  !rev.equals(l_rev = getLastRevision(docid)) ) {
        throw new AccessionNumberException
                  ("Last revision number is "+ l_rev);
      }
    }
  }

  /**
   * Generate an Accession Number of form <sidecode>.<uniqueid>.<revisionid>
   * where <revisionid> is always 1.
   *
   * @param docid <sitecode>.<uniqueid> part of Accession Number
   * @param action INSERT, UPDATE or DELETE action
   */
  public String generate(String docid, String action) 
                throws AccessionNumberException, SQLException 
  {
    return generate(docid, "1", action);
  }

  /**
   * Generate an Accession Number of form <sidecode>.<uniqueid>.<revisionid>
   *
   * @param docid <sitecode>.<uniqueid> part of Accession Number
   * @param rev <revisionid> of Accession Number
   * @param action INSERT, UPDATE or DELETE action
   */
  public String generate(String docid, String rev, String action) 
                throws AccessionNumberException, SQLException 
  {
    try {
      // INSERT
      if ( action.equals("INSERT")) {

        // get a new docid
        if ( docid == null ) {
          String sitecode = getSitecode();
          String uniqueid = getUniqueID(sitecode);

          return sitecode + sep + uniqueid;

        // docid is used; throw an exception to prevent the insertion
        } else if ( accNumberUsed(docid) ) {
          throw new AccessionNumberException
                    ("Accession number " + docid + " is already in use.");
        // rev is <> 1; throw an exception to prevent the insertion
        } else if ( !rev.equals("1") ) {
          throw new AccessionNumberException
                    ("Revision number must be 1.");
        // docid is not used and rev=1; return it
        } else {
          return docid;
        }

      // UPDATE or DELETE
      } else if ( action.equals("UPDATE") || action.equals("DELETE")) {

        // Accession# is not provided; throw an exception to prevent the action
        if ( docid == null ) {
          throw new AccessionNumberException("Accession number is required.");

        // Accession# is not current (not in xml_documents); throw an exception
        } else if ( !accNumberIsCurrent(docid) ) {
          throw new AccessionNumberException
          ("Document not found for Accession number " + docid);

        // Revision number is not the recent one; throw an exception
        } else if ( !rev.equals(getLastRevision(docid)) ) {
          throw new AccessionNumberException
          ("Revision number must be the recent.");

        // Accession# is current (present in xml_documents); return it
        } else {
          return docid;
        }
      }
 
    } catch (SQLException e) {
      throw new SQLException
      ("Error on AccessionNumber.generate(): " + e.getMessage());
    }    
 
    // never comes here
    throw new
    AccessionNumberException("Fatal Error in accession number generation.");
  }
  
  // get local sitecode
  private String getSitecode()
  {
    return this.sitecode;
  }

  // get Unique ID from DB sequence
  private String getUniqueID (String sitecode) throws SQLException
  {
    String uniqueid;
    String sysdate = dbAdapter.getDateTimeFunction();
    
    try {
      PreparedStatement pstmt;
      pstmt = conn.prepareStatement
              ("INSERT INTO accession_number (site_code, date_created) " +
               "VALUES (?, " + sysdate + ")");
      pstmt.setString(1, sitecode);
      pstmt.execute();
      pstmt.close();
      
      uniqueid = "" + dbAdapter.getUniqueID(conn, "accession_number");

    } catch (SQLException e) {
      throw new 
      SQLException("Error on AccessionNumber.getUniqueID(): "+e.getMessage());
    }

    return uniqueid;
  }

  /** check for existence of Accesssion Number xml_acc_numbers table */
  public boolean accNumberUsed ( String accNumber )
                  throws SQLException {
        
    boolean hasAccNumber = false;
        
    try {
      PreparedStatement pstmt;
      pstmt = conn.prepareStatement(
                "SELECT 'x' FROM xml_documents " + 
                "WHERE docid = ? " +
                "UNION " +
                "SELECT 'x' FROM xml_revisions " +
                "WHERE docid = ?");
      pstmt.setString(1,accNumber);
      pstmt.setString(2,accNumber);
      pstmt.execute();
      ResultSet rs = pstmt.getResultSet();
      hasAccNumber = rs.next();
      pstmt.close();
            
    } catch (SQLException e) {
      throw new SQLException
      ("Error on AccessionNumber.accNumberUsed(accNumber): " + e.getMessage());
    }    
        
    return hasAccNumber;
  }    
    
  // check for existence of Accesssion Number in xml_documents table
  private boolean accNumberIsCurrent(String accNumber) throws SQLException {
        
    boolean hasCurrentAccNumber = false;
        
    try {
      PreparedStatement pstmt;
      pstmt = conn.prepareStatement(
                "SELECT 'x' FROM xml_documents " + 
                "WHERE docid = ?");
      pstmt.setString(1, accNumber);
      pstmt.execute();
      ResultSet rs = pstmt.getResultSet();
      hasCurrentAccNumber = rs.next();
      pstmt.close();
            
    } catch (SQLException e) {
      throw new SQLException(
          "Error on AccessionNumber.accNumberIsCurrent(String accNumber): " +
          e.getMessage());
    }    
  
    return hasCurrentAccNumber;
  }    

  // get the recent revision number for docid
  private String getLastRevision(String docid) throws SQLException
  {
    String rev = "";
    
    try {
      PreparedStatement pstmt;
      pstmt = conn.prepareStatement
              ("SELECT rev FROM xml_documents WHERE docid='" + docid + "'");
      pstmt.execute();

      ResultSet rs = pstmt.getResultSet();
      boolean hasRow = rs.next();
      rev = rs.getString(1);
      pstmt.close();
      
    } catch (SQLException e) {
      throw new SQLException(
      "Error on AccessionNumber.getLastRevision(): " + e.getMessage());
    }

    return rev;
  }

  // get the next revision number for docid
  private String getNextRevision(String docid) throws SQLException
  {
    String rev = "";
    
    try {
      PreparedStatement pstmt;
      pstmt = conn.prepareStatement
              ("SELECT rev+1 FROM xml_documents WHERE docid='" + docid + "'");
      pstmt.execute();

      ResultSet rs = pstmt.getResultSet();
      boolean hasRow = rs.next();
      rev = rs.getString(1);
      pstmt.close();
      
    } catch (SQLException e) {
      throw new SQLException(
      "Error on AccessionNumber.getNextRevision(): " + e.getMessage());
    }

    return rev;
  }
  
  /**
   * returns the docid encoded in this accession number
   */
  public String getDocid() {

    return this.docid;
  }

  /**
   * returns the revision number encoded in this accession number
   */
  public String getRev()
  {
    return this.rev;
  }
  
}
