/**
 *  '$RCSfile: DBTransform.java,v $'
 *    Purpose: A Class that transforms an XML text document
 *             into a another type using XSL
 *  Copyright: 2000 Regents of the University of California and the
 *             National Center for Ecological Analysis and Synthesis
 *    Authors: Matt Jones
 *    Release: 1.0.3
 *
 *   '$Author: jones $'
 *     '$Date: 2001/09/14 23:00:54 $'
 * '$Revision: 1.13 $'
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package edu.ucsb.nceas.metacat;

import java.io.*;
import java.net.URL;
import java.net.MalformedURLException;
import java.sql.*;
import java.util.Stack;

import oracle.xml.parser.v2.XSLStylesheet;
import oracle.xml.parser.v2.XSLException;
import oracle.xml.parser.v2.XMLParseException;
import oracle.xml.parser.v2.XSLProcessor;
import oracle.xml.parser.v2.XMLDocument;
import oracle.xml.parser.v2.DOMParser;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;

/** 
 * A Class that transforms XML documents utitlizing XSL style sheets
 */
public class DBTransform {

  private Connection	conn = null;
  private MetaCatUtil   util = null;
  private String 	configDir = null;
  private String	defaultStyle = null;

  /**
   * construct a DBTransform instance.
   *
   * Generally, one calls transformXMLDocument() after constructing the instance
   *
   * @param conn the database connection from which to lookup the public ids
   */
  public DBTransform( Connection conn ) 
                  throws IOException, 
                         SQLException, 
                         ClassNotFoundException
  {
    this.conn = conn;
    util = new MetaCatUtil();
    configDir = util.getOption("config-dir");
    defaultStyle = util.getOption("default-style");
  }
  
  /**
   * Transform an XML document using the stylesheet reference from the db
   *
   * @param doc the document to be transformed
   * @param sourcetype the document type of the source
   * @param targettype the target document type
   * @param qformat the name of the style set to use
   * @param pw the PrintWriter to which output is printed
   */
  public void transformXMLDocument(String doc, String sourcetype, 
                String targettype, String qformat, PrintWriter pw) {
    
    // Look up the stylesheet for this type combination
    String xsl_system_id = getStyleSystemId(qformat, sourcetype, targettype);

    if (xsl_system_id != null) {
      // Create a stylesheet from the system id that was found
      try {
        XSLStylesheet style = new XSLStylesheet(new URL(xsl_system_id), null);
        DOMParser dp = new DOMParser();
        dp.setValidationMode(false);
        dp.parse((Reader)(new StringReader(doc)));
        new XSLProcessor().processXSL(style, dp.getDocument(), pw);
      } catch (Exception e) {
        pw.println(xsl_system_id + "Error transforming document in " +
                   "DBTransform.transformXMLDocument: " +
                   e.getMessage());
      }
    } else {
      // No stylesheet registered form this document type, so just return the 
      // XML stream we were passed
      pw.print(doc);
    }
  }

  /**
   * Lookup a stylesheet reference from the db catalog
   *
   * @param qformat    the named style-set format
   * @param sourcetype the document type of the source
   * @param targettype the document type of the target
   */
  public String getStyleSystemId(String qformat, String sourcetype, 
                String targettype) {
    String systemId = null;

    if ((qformat == null) || (qformat.equals("html"))) {
      qformat = defaultStyle;
    }

    // Load the style-set map for this qformat into a DOM
    try {
      String filename = configDir + "/" + qformat + ".xml";       
      util.debugMessage("Trying style-set file: " + filename);

      DOMParser dp = new DOMParser();
      dp.setValidationMode(false);
      dp.parse((Reader)(new FileReader(filename)));
      Document doc = dp.getDocument();
      Element root = doc.getDocumentElement();
      String styleName = root.getAttribute("name");
      util.debugMessage("Root style-set element is: " + styleName);
      
      Element currentElement = (Element)root.getFirstChild();
      systemId = ((Node)currentElement.getFirstChild()).getNodeValue();
      util.debugMessage("Default is: " + systemId);

      while ((currentElement = 
              (Element)currentElement.getNextSibling()) != null) {
        String tagName = currentElement.getTagName();
        util.debugMessage("Processing element: " + tagName);
        if (tagName.equals("doctype")) {
          String doctype = currentElement.getAttribute("publicid");
          util.debugMessage("Processing publicid: " + doctype);
          util.debugMessage("Comparing to source: " + sourcetype);
          if (doctype.equals(sourcetype)) {
            Element currentChild = (Element)currentElement.getFirstChild();
            while (currentChild != null) {
              String target = currentChild.getAttribute("publicid");
              util.debugMessage("Processing target publicid: " + target);
              util.debugMessage("Comparing to target: " + targettype);
              if (target.equals(targettype)) {
                Node styleText = currentChild.getFirstChild();
                systemId = styleText.getNodeValue();
                break; 
              }
              currentChild = (Element)currentChild.getNextSibling();
            }
          }
        }
      }
    } catch (IOException ioe) {
      util.debugMessage("Caught IOException while opening style-set config.");
    } catch (XMLParseException xpe) {
      util.debugMessage("Error parsing style-set file");
    } catch (SAXException se) {
      util.debugMessage("SAX error parsing style-set file");
    }

    // Return the system ID for this particular source document type
    return systemId;
  }

  /**
   * Lookup a stylesheet reference from the db catalog
   *
   * @param objecttype the type of the object we want to retrieve
   * @param sourcetype the document type of the source
   * @param targettype the document type of the target
   */
  public String getSystemId(String objecttype, String sourcetype, 
                String targettype) {

    // Look up the System ID of a particular object
    PreparedStatement pstmt;
    String the_system_id = null;
    try {
      pstmt =
        conn.prepareStatement("SELECT system_id " +
                "FROM xml_catalog " +
                "WHERE entry_type = ? " +
                "AND source_doctype = ? " +
                "AND target_doctype = ? ");
      // Bind the values to the query
      pstmt.setString(1, objecttype);
      pstmt.setString(2, sourcetype);
      pstmt.setString(3, targettype);
      pstmt.execute();
      try {
        ResultSet rs = pstmt.getResultSet();
        try {
          boolean tableHasRows = rs.next();
          if (tableHasRows) {
            try {
              the_system_id = rs.getString(1);
            } catch (SQLException e) {
              System.out.println("Error with getString in " + 
                                 "DBTransform.getSystemId: " + e.getMessage());                
            }
          } else {
            the_system_id = null; 
          }
        } catch (SQLException e) {
          System.err.println("Error with next in DBTransform.getSystemId: " + 
                              e.getMessage());
          return ("Error with next: " + e.getMessage());
        }
      } catch (SQLException e) {
        System.err.println("Error with getrset in DBTransform.getSystemId: " + 
                            e.getMessage());
        return ("Error with getrset: " + e.getMessage());
      }
      pstmt.close();
    } catch (SQLException e) {
      System.err.println("Error getting id in DBTransform.getSystemId: " + 
                          e.getMessage());
      return ("Error getting id in DBTransform.getSystemId:: " + 
               e.getMessage());
    }
    return the_system_id;
  }

  /**
   * the main routine used to test the transform utility.
   *
   * Usage: java DBTransform
   */
  static public void main(String[] args) {
     
     if (args.length > 0)
     {
        System.err.println("Wrong number of arguments!!!");
        System.err.println("USAGE: java DBTransform");
        return;
     } else {
        try {
                    
          // Open a connection to the database
          MetaCatUtil   util = new MetaCatUtil();
          Connection dbconn = util.openDBConnection();

          // Create a test document
          StringBuffer testdoc = new StringBuffer();
          testdoc.append("<?xml version=\"1.0\"?>");
          testdoc.append("<eml-dataset><metafile_id>NCEAS-0001</metafile_id>");
          testdoc.append("<dataset_id>DS001</dataset_id>");
          testdoc.append("<title>My test doc</title></eml-dataset>");

          // Transform the document to the new doctype
          DBTransform dbt = new DBTransform(dbconn);
          dbt.transformXMLDocument(testdoc.toString(), 
                                   "-//NCEAS//eml-dataset//EN", 
                                   "-//W3C//HTML//EN", 
                                   "knb",
                                   new PrintWriter(System.out));

        } catch (Exception e) {
          System.err.println("EXCEPTION HANDLING REQUIRED");
          System.err.println(e.getMessage());
          e.printStackTrace(System.err);
        }
     }
  }
  
  private void dbg(int position) {
    System.err.println("Debug flag: " + position);
  }

}
